//****************************************************************
//
//    Class entry
//
//    Description:
//	Entry is a class for implementing an entry in a symbol
//	table.  Each entry is composed of a string pointer
// 	and a value.  (See class value.)
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 18, 1995
//
//****************************************************************

/*==========================================================================
				PRIVATE :
				--------

    char *name;		// name of the entry
    value data;		// value of the entry


				FUNCTIONS :
				----------

constructors
destructor

get_name : 		returns the pointer to the name 
get_value : 		returns the value component of the entry
set_name(n) : 		sets the name of the entry to the string pointed to by n
set_value_str(n) : 	sets the value of the entry to the string pointed to by
			n
set_value_num(n) : 	sets the value of the entry to the number n
set_value_dense(n) :	sets the value of the entry to the dense matrix n
set_value_sparse(n) : 	sets the value of the entry to the sparse matrix n
operator =

==========================================================================*/


#ifndef ENTRY_H
#define ENTRY_H

#include "value.h"

class entry {

  public:
    // Constructors
    entry();
    entry(char *);
    entry(char *, value);
    entry(const entry &);

    // Selectors    
    char * get_name() const;
    value get_value() const;

    // Mutators
    void set_name(char *);
    void set_value_str(char *);
    void set_value_num(double);
    void set_value_dense(dense *);
    void set_value_sparse(sparse *);
    void free();

    // Operator
   entry operator = (const entry &);


  private:
    // Data
    char *name;		// name of the entry
    value data;		// value of the entry

}; // end class entry

#endif
