// ****************************************************************
//
//    Builtin Operators
//
//    Description:
//	Contains all of the builtin operators.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 18, 1995
//
// ****************************************************************

#include "value.h"
#include "operators.h"

extern int yyerror(const char *);

/*

  if (left->type == NUM) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

*/

/*============================================================================
#				POW					      
#				---					      	
# in entry  : 	left  ->
#		right -> pow
#
# in return :	a value
#
# purpose   :	it calculates left ^ right
#
# details   :	it searchs the type of the entries and verifies that there are
#		no errors (ex: matrices must be raised to integer powers), then
#	 	it calls the right function.
#	left=NUM and right=NUM : checks that if left is neg, right is int
#				 returns left * right by calling pow(left,right)
#  	left=DENSE right=NUM   : checks if right is an integer
#				 checks if left is square
#		  	    	 returns left * right (if OK) by making a loop	
# 	left=SPARSE right=NUM :  idem
#	else 			 returns an error 
#
============================================================================*/

value * pow (value *left, value *right, int &err) {
  
  if (left->type == NUM) {
    if (right->type == NUM) {
      if (left->num < 0  && right->num != int(right->num)) {
        yyerror ("Error, negative numbers must be raised to integer powers.");
        err = 1;
	return new value (0);
      } // end if
      return new value (pow(left->num,right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      if (right->num != int(right->num)) {
        yyerror ("Error, matrices must be raised to integer powers.");
        err = 1;
	return new value (0);
      } // end if
      if (left->dmat->get_nrow() != left->dmat->get_ncol()) {
        yyerror ("Error, matrices raised to a power must be square.");
        err = 1;
	return new value (0);
      } // end if
      dense result(left->dmat->get_nrow(), left->dmat->get_ncol());
      result.make_identity();
      int i;
      for (i = 1; i <= right->num; i += 1)
	result = result * *left->dmat;
      return new value (new dense (result));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      if (right->num != int(right->num)) {
        yyerror ("Error, matrices must be raised to integer powers.");
        err = 1;
	return new value (0);
      } // end if
      if (left->smat->get_nrow() != left->smat->get_ncol()) {
        yyerror ("Error, matrices raised to a power must be square.");
        err = 1;
	return new value (0);
      } // end if
      sparse result;
      result.init_identity(left->smat->get_nrow());
      int i;
      for (i = 1; i <= right->num; i += 1)
	result = ((*left->smat) * (result));
      return new value (new sparse (result));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator '");
  return left;

} // end pow

/*============================================================================
#				EPOW					      
#				----					      	
# in entry  : 	left  ->
#		right -> pow
#
# in return :	a value
#
# purpose   :	it calculates left ^ right element by element. 
#
# details   :	it searchs the type of the entries and verifies that there are
#		no errors (ex: negative numbers must be raised to integer 
#		powers), then it calls the right function.
#	left=NUM and right=NUM : checks that if right is neg, right is int.
#				 returns left * right by calling pow(left,right)
#  	left=DENSE right=NUM   : for each nij=left[i,j], calls mij=pow(nij,left)
#				 and returns the matrix R with R[i,j]=mij. 
# 	left=SPARSE right=NUM :  idem as pow
#
============================================================================*/


value * epow (value *left, value *right, int &err) {
  
  int i,j;

  if (left->type == NUM) {
    if (right->type == NUM) {
      if (left->num < 0  && right->num != int(right->num)) {
        yyerror ("Error, negative numbers must be raised to integer powers.");
        err = 1;
	return new value (0);
      } // end if
      return new value (pow(left->num,right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      dense *result = new dense (*left->dmat);
      for (i = 1; i <= left->dmat->get_nrow(); i += 1)
        for (j = 1; j <= left->dmat->get_ncol(); j += 1)
          result->set_element(i,j,pow(left->dmat->get_element(i,j),right->num));
      return new value (result);
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      if (right->num != int(right->num)) {
        yyerror ("Error, matrices must be raised to integer powers.");
        err = 1;
	return new value (0);
      } // end if
      if (left->smat->get_nrow() != left->smat->get_ncol()) {
        yyerror ("Error, matrices raised to a power must be square.");
        err = 1;
	return new value (0);
      } // end if
      sparse result;
      result.init_identity(left->smat->get_nrow());
      int i;
      for (i = 1; i <= right->num; i += 1)
	result = (*left->smat * result);
      return new value (new sparse (result));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator '");
  return left;

} // end epow

/*============================================================================
#				TRANSP					      
#				---					      	
# in entry  : 	left  -> a value
#		
# in return :	a value (same type as in entry)
#
# details   :	it searchs the type of the entry and calculates the transpose of # 		left
#  	left=DENSE   :	calls the function transp of the class dense.c
#	left=SPARSE  :	calls the function transp of the class sparse.c
#	else         :	returns left

 ============================================================================*/

value * transp (value *left, int & err) {
  // Compute the logical not of left

  if (left->type == DENSE) 
    return new value (new dense(transp(*left->dmat)));

  if (left->type == SPARSE)
    return new value (new sparse(transp(*left->smat)));

  err = 1;
  yyerror ("Error, incompatible types for operator '");
  return left;

} // end transp

/*============================================================================
#				PLUS					      
#				----					      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :	n+n=n, n+d=d, n+s=d, d+n=d, d+d=d, d+s=d, s+n=d, s+d=d, s+s=s
#		with n=NUM, d=DENSE, s=SPARSE
#
# details   :	It add two values (not necessary of the same type) and returns
# 		the result.
#	left=NUM and right=NUM : returns left + right
#  	left=DENSE right=NUM   : returns M/ M[i,j] = right + left[i,j]
# 	left=SPARSE right=NUM  : idem
#	left=NUM right=DENSE   : returns M/ M[i,j] = left + right[i,j]
#	left=DENSE right=DENSE : checks if left and right have the same
#				 dimensions
#				 calls the operator + of the class dense.c
#	left=SPARSE right=DENSE: for the moment there will be a bus error! But 
# 		I have changed and it should work.
#	left=SPARSE right=SPARSE: checks if left and right have the same
#				  dimensions
				  calls the operator + of the class sparse.c
#	left=NUM right=SPARSE  : returns M/M[i,j] = left + right[i,j]
#	left=DENSE right=SPARSE: checks if left and right have the same dim 
#				 it densifies right and calls the operator +
#				 of the class dense.c
#
# Bus errors occur because the function dense(sparse mat) doesn't seem to work!
#
============================================================================*/

value * plus (value *left, value *right, int &err) {
  // Add the two values and return the result.

  if (left->type == NUM) {
    if (right->type == NUM) {
      return new value ((left->num + right->num));
    } // end if
    if (right->type == DENSE) {
      return new value (new dense (left->num + *right->dmat));
    } // end if
    if (right->type == SPARSE) {
      return new value(  new dense(left->num + dense(*right->smat )));
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat + right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->dmat->get_nrow() == right->dmat->get_nrow() && left->dmat->get_ncol() == right->dmat->get_ncol())
        return new value (new dense (*left->dmat + *right->dmat));
      else {
	yyerror ("Error, imcompatible dimensions for operator +");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == SPARSE) {
      if (left->dmat->get_nrow() == right->smat->get_nrow() && left->dmat->get_ncol() == right->smat->get_ncol())
        return new value (new dense (*left->dmat + dense(*right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator +");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new dense (dense(*left->smat) + right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->smat->get_nrow() == right->dmat->get_nrow() && left->smat->get_ncol() == right->dmat->get_ncol())
        return new value (new dense (*left->dmat + dense(*right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator +");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == SPARSE) {
      if (left->smat->get_nrow() == right->smat->get_nrow() && left->smat->get_ncol() == right->smat->get_ncol())
        return new value (new sparse (*left->smat + *right->smat));
      else {
	yyerror ("Error, imcompatible dimensions for operator +");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == STRING) {
       // do nothing
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
      // do nothing
    } // end if
    if (right->type == DENSE) {
      // do nothing
    } // end if
    if (right->type == SPARSE) {
      // do nothing
    } // end if
    if (right->type == STRING) {
      char *str = new char [strlen(left->str) + strlen(right->str) + 1];
      strcpy (str,left->str);
      strcat (str,right->str);
      return new value (str);
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator +");
  return left;

} // end plus

/*============================================================================
#				MINUS					      
#				-----				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :	n-n=n, n-d=d, n-s=d, d-n=d, d-d=d, d-s=d, s-n=d, s-d=d, s-s=s
#		with n=NUM, d=DENSE, s=SPARSE
#
# details   :	It substracts two values (not necessary of the same type) and
# 		returns the result.
#	left=NUM and right=NUM : returns left - right
#  	left=DENSE right=NUM   : returns M/ M[i,j] = left[i,j] - right
# 	left=SPARSE right=NUM  : idem
#	left=NUM right=DENSE   : returns M/ M[i,j] = left - right[i,j]
#	left=DENSE right=DENSE : checks if left and right have the same
#				 dimensions
#				 calls the operator - of the class dense.c
#	left=SPARSE right=DENSE: for the moment there will be a bus error!
#	but I have made changes and it should work
#	left=SPARSE right=SPARSE: checks if left and right have the same
#				  dimensions
				  calls the operator - of the class sparse.c
#	left=NUM right=SPARSE  : returns M/M[i,j] = left - right[i,j]
#	left=DENSE right=SPARSE: checks if left and right have the same dim 
#				 it densifies right and calls the operator -
#				 of the class dense.c
#
# Bus errors occur because peter densified the dense matrix and not the sparse 
# one
#
============================================================================*/



value * minus (value *left, value *right, int &err) {
  // Subtract the two values and return the result.

  if (left->type == NUM) {
    if (right->type == NUM) {
      return new value ((left->num - right->num));
    } // end if
    if (right->type == DENSE) {
      return new value (new dense (left->num - *right->dmat));
    } // end if
    if (right->type == SPARSE) {
      return new value (new dense (left->num - dense(*right->smat)));
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat - right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->dmat->get_nrow() == right->dmat->get_nrow() && left->dmat->get_ncol() == right->dmat->get_ncol())
        return new value (new dense (*left->dmat - *right->dmat));
      else {
	yyerror ("Error, imcompatible dimensions for operator -");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == SPARSE) {
      if (left->dmat->get_nrow() == right->smat->get_nrow() && left->dmat->get_ncol() == right->smat->get_ncol())
        return new value (new dense (*left->dmat - dense(*right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator -");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new dense (dense(*left->smat) - right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->smat->get_nrow() == right->dmat->get_nrow() && left->smat->get_ncol() == right->dmat->get_ncol())
        return new value (new dense (*left->dmat - dense(*right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator -");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == SPARSE) {
      if (left->smat->get_nrow() == right->smat->get_nrow() && left->smat->get_ncol() == right->smat->get_ncol())
        return new value (new sparse (*left->smat - *right->smat));
      else {
	yyerror ("Error, imcompatible dimensions for operator -");
	err = 1;
	return left;
      }
    } // end if
    if (right->type == STRING) {
       // do nothing
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
      // do nothing
    } // end if
    if (right->type == DENSE) {
      // do nothing
    } // end if
    if (right->type == SPARSE) {
      // do nothing
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator -");
  return left;

} // end minus

/*============================================================================
#				MULT					      
#				----				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :	n*n=n, n*d=d, n*s=s, d*n=d, d*d=d, d*s=d, s*n=s, s*d=d, s*s=s
#		with n=NUM, d=DENSE, s=SPARSE
#
# details   :	It multiplies two values (not necessary of the same type) and
# 		returns the result.
#	left=NUM and right=NUM : returns left * right
#  	left=DENSE right=NUM   : returns M/ M[i,j] = left[i,j] * right
# 	left=SPARSE right=NUM  : idem
#	left=NUM right=DENSE   : returns M/ M[i,j] = left * right[i,j]
#	left=DENSE right=DENSE : checks if left and right have the same
#				 dimensions
#				 calls the operator * of the class dense.c
#	left=SPARSE right=DENSE: checks if left and right have the same dim 
#				 it densifies left and calls the operator *
#				 of the class dense.c
#	left=SPARSE right=SPARSE: checks if left and right have the same
#				  dimensions
				  calls the operator * of the class sparse.c
#	left=NUM right=SPARSE  : returns M/M[i,j] = left * right[i,j]
#	left=DENSE right=SPARSE: checks if left and right have the same dim 
#				 it densifies right and calls the operator *
#				 of the class dense.c
#
============================================================================*/



value * mult (value *left, value *right, int & err) {
  // Multiply the two values and return the result.


  if (left->type == NUM) {
    if (right->type == NUM) {
      return new value ((left->num * right->num));
    } // end if
    if (right->type == DENSE) {
      return new value (new dense (left->num * *right->dmat));
    } // end if
    if (right->type == SPARSE) {
      return new value (new sparse (left->num * *right->smat));
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat * right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() == right->dmat->get_nrow())
        return new value (new dense ((*left->dmat) * (*right->dmat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator *");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->dmat->get_ncol() == right->smat->get_nrow())
        return new value (new dense ((*left->dmat) * dense(*right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator *");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new sparse (*left->smat * right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->smat->get_ncol() == right->dmat->get_nrow())
        return new value (new dense (dense(*left->smat) * *right->dmat));
      else {
	yyerror ("Error, imcompatible dimensions for operator *");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->smat->get_ncol() == right->smat->get_nrow())
        return new value (new sparse (*left->smat * *right->smat));
      else {
	yyerror ("Error, imcompatible dimensions for operator *");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
       // do nothing
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
      // do nothing
    } // end if
    if (right->type == DENSE) {
      // do nothing
    } // end if
    if (right->type == SPARSE) {
      // do nothing
    } // end if
    if (right->type == STRING) {
      // do nothing
    } // end if
  } // end if


  err = 1;
  yyerror ("Error, incompatible types for operator *");
  return left;

} // end mult

/*============================================================================
#				EMULT					      
#				-----				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   like mult!
#
# details   :	It multiplies two values (not necessary of the same type)
#		element by element and returns the result.
#	left=NUM and right=NUM : returns left * right
#  	left=DENSE right=NUM   : returns M/ M[i,j] = left[i,j] * right
# 	left=SPARSE right=NUM  : idem result=sparse
#	left=NUM right=DENSE   : returns M/ M[i,j] = left * right[i,j]
#	left=DENSE right=DENSE : checks if left and right have the same
#				 dimensions
#				 calls the operator .* of the class dense.c
#	left=SPARSE right=DENSE: checks if left and right have the same dim 
#				 it densifies left and calls the operator .*
#				 of the class dense.c
#	left=SPARSE right=SPARSE: checks if left and right have the same dim 
#				 it densifies left and right and calls the 
#				 operator .* of the class dense.c
#				 of the class dense.c
#	left=NUM right=SPARSE  : returns M/M[i,j] = left * right[i,j]
#	left=DENSE right=SPARSE: checks if left and right have the same dim 
#				 it densifies right and calls the operator *
#				 of the class dense.c
#
============================================================================*/



value * emult (value *left, value *right, int & err) {
  // Multiply the two values element by element


  if (left->type == NUM) {
    if (right->type == NUM) {
      return new value ((left->num * right->num));
    } // end if
    if (right->type == DENSE) {
      return new value (new dense (left->num * *right->dmat));
    } // end if
    if (right->type == SPARSE) {
      return new value (new sparse (left->num * *right->smat));
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat * right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() == right->dmat->get_ncol()  &&
	  left->dmat->get_nrow() == right->dmat->get_nrow())
        return new value (new dense (emult(*left->dmat , *right->dmat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator .*");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->dmat->get_ncol() == right->smat->get_ncol()  &&
	  left->dmat->get_nrow() == right->smat->get_nrow())
        return new value (new dense (emult(*left->dmat , dense(*right->smat))));
      else {
	yyerror ("Error, imcompatible dimensions for operator .*");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new sparse (*left->smat * right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->smat->get_ncol() == right->dmat->get_ncol()  &&
	  left->smat->get_nrow() == right->dmat->get_nrow())
        return new value (new dense (emult(dense(*left->dmat) , *right->dmat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator .*");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->smat->get_ncol() == right->smat->get_ncol()  &&
	  left->smat->get_nrow() == right->smat->get_nrow())
        return new value (new dense (emult(dense(*left->dmat) ,dense(*right->dmat))));
      else {
	yyerror ("Error, imcompatible dimensions for operator .*");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator .*");
  return left;

} // end emult

/*============================================================================
#				DIV					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   like mult!
#
# details   :	It divides two values (not necessary of the same type)
#		 and returns the result.
#		In each case, it verifies that right is a number and 
#		isn't equal to 0. Otherwise, it returns an error.
#		If everything is OK, it calls the operator / of the good class
#
============================================================================*/



value * div (value *left, value *right, int &err) {
  // devide the two values and return the result.

  if (left->type == NUM) {
    if (right->type == NUM) {
      if (right->num == 0) {
	// Do nothing, just let it devide it out, and get infinity.
        //err = 1;
	//yyerror ("Error, devision by zero.");
	//return left;
      } // end if
      return new value ((left->num / right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat / right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new sparse (*left->smat / right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator /");
  return left;

} // end div


/*============================================================================
#				EDIV					      
#				----				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   like mult!
#
# details   :	It divides two values (not necessary of the same type)
#		element by element and returns the result.
#		In each case, it verifies that right is a number and 
#		isn't equal to 0. Otherwise, it returns an error.
#		If everything is OK, it calls the operator / of the good class
#
============================================================================*/



value * ediv (value *left, value *right, int &err) {
  // devide the two values element by element and return the result.


  if (left->type == NUM) {
    if (right->type == NUM) {
      if (right->num == 0) {
        err = 1;
	yyerror ("Error, devision by zero.");
	return left;
      } // end if
      return new value ((left->num / right->num));
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM) {
      return new value (new dense (*left->dmat / right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() == right->dmat->get_ncol()  &&
	  left->dmat->get_nrow() == right->dmat->get_nrow())
        return new value (new dense (ediv(*left->dmat , *right->dmat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator ./");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->dmat->get_ncol() == right->smat->get_ncol()  &&
	  left->dmat->get_nrow() == right->smat->get_nrow())
        return new value (new dense (ediv(*left->dmat , dense(*right->smat))));
      else {
	yyerror ("Error, imcompatible dimensions for operator ./");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == SPARSE) {
    if (right->type == NUM) {
      return new value (new sparse (*left->smat / right->num));
    } // end if
    if (right->type == DENSE) {
      if (left->smat->get_ncol() == right->dmat->get_ncol()  &&
	  left->smat->get_nrow() == right->dmat->get_nrow())
        return new value (new dense (ediv(dense(*left->dmat) , *right->smat)));
      else {
	yyerror ("Error, imcompatible dimensions for operator ./");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == SPARSE) {
      if (left->smat->get_ncol() == right->smat->get_ncol()  &&
	  left->smat->get_nrow() == right->smat->get_nrow())
        return new value (new dense (ediv(dense(*left->dmat) , dense(*right->smat))));
      else {
	yyerror ("Error, imcompatible dimensions for operator ./");
	err = 1;
	return left;
      } // end else
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if

  if (left->type == STRING) {
    if (right->type == NUM) {
    } // end if
    if (right->type == DENSE) {
    } // end if
    if (right->type == SPARSE) {
    } // end if
    if (right->type == STRING) {
    } // end if
  } // end if




  if (left->type == DENSE)
    if (right->type == NUM)
      return new value (new dense (*left->dmat / right->num));

  err = 1;
  yyerror ("Error, incompatible types for operator ./");
  return left;

} // end ediv


/*============================================================================
#				AND					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes the logical and of the two values (not necessary
# 		of the same type)
#		 and returns the result.
#		It calls the operator && of the good class
#
============================================================================*/



value * _and (value *left, value *right, int & err) {
  // Compute the logical and of the two values.

  if (left->type == NUM) {
    if (right->type == NUM) {
      if (left->num == 0)
        return new value ((0));
      if (right->num == 0)
	return new value ((0));
      return new value ((1));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num && *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat && right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator &&");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator &&");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat && *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator &&");
  return left;

} // end and

/*============================================================================
#				NOT					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes the logical not of left and returns the result.
#		It calls the operator ! of the good class
#
============================================================================*/



value * _not (value *left, int & err) {
  // Compute the logical not of left

  if (left->type == NUM) 
    if (left->num == (0))
      return new value ((1));
    else
      return new value ((0));

  if (left->type == DENSE) 
    return new value (new dense(!(*left->dmat)));

  err = 1;
  yyerror ("Error, incompatible types for operator !");
  return left;

} // end not

/*============================================================================
#				OR					      
#				--				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes the logical or of the two values (not necessary
# 		of the same type)
#		and returns the result.
#		It calls the operator || of the good class
#
============================================================================*/



value * _or (value *left, value *right, int & err) {
  // Compute the logical or of the two values.

  if (left->type == NUM) {
    if (right->type == NUM) {
      if (left->num != 0)
        return new value ((1));
      if (right->num != 0)
	return new value ((1));
      return new value ((0));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num || *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat || right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator ||");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator ||");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat || *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator ||");
  return left;

} // end or

/*============================================================================
#				LT					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left<right 
#		and returns the result.
#		It calls the operator < of the good class
#
============================================================================*/



value * lt (value *left, value *right, int & err) {
  // Compute left < right.

  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num < right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num < *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat < right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator <");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator <");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat < *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator <");
  return left;

} // end lt

/*============================================================================
#				LE					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left<=right 
#		and returns the result.
#		It calls the operator <= of the good class
#
============================================================================*/



value * le (value *left, value *right, int & err) {
  // Compute left <= right.


  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num <= right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num <= *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat <= right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator <=");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator <=");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat <= *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator <=");
  return left;

} // end le

/*============================================================================
#				EQ					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left==right 
#		and returns the result.
#		It calls the operator == of the good class
#
============================================================================*/



value * eq (value *left, value *right, int & err) {
  // Compute left == right.

  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num == right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num == *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat == right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator ==");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator ==");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat == *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator ==");
  return left;

} // end eq

/*============================================================================
#				NE					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left!=right 
#		and returns the result.
#		It calls the operator != of the good class
#
============================================================================*/



value * ne (value *left, value *right, int & err) {
  // Compute left != right.

  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num != right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num != *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat != right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator !=");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator !=");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat != *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator !=");
  return left;

} // end ne

/*============================================================================
#				GE					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left>=right 
#		and returns the result.
#		It calls the operator >= of the good class
#
============================================================================*/



value * ge (value *left, value *right, int & err) {
  // Compute left >= right.

  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num >= right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num >= *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat >= right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator >=");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator >=");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat >= *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator >=");
  return left;

} // end ge

/*============================================================================
#				GT					      
#				---				      	
# in entry  : 	left  -> a value
#		right -> a value
#
# in return :   
#
# details   :	It computes left>right 
#		and returns the result.
#		It calls the operator > of the good class
#
============================================================================*/



value * gt (value *left, value *right, int & err) {
  // Compute left > right.

  if (left->type == NUM) {
    if (right->type == NUM) {
        return new value ((left->num > right->num));
    } // end if
    if (right->type == DENSE) 
      return new value (new dense(left->num > *right->dmat));
  } // end if

  if (left->type == DENSE) {
    if (right->type == NUM)
      return new value (new dense(*left->dmat > right->num));
    if (right->type == DENSE) {
      if (left->dmat->get_ncol() != right->dmat->get_ncol()) {
	yyerror ("Error, imcompatible dimensions for operator >");
	err = 1;
	return left;
      } // end if
      if (left->dmat->get_nrow() != right->dmat->get_nrow()) {
	yyerror ("Error, imcompatible dimensions for operator >");
	err = 1;
	return left;
      } // end if
      return new value (new  dense (*left->dmat > *right->dmat));
    } // end if
  } // end if

  err = 1;
  yyerror ("Error, incompatible types for operator >");
  return left;

} // end gt

/*============================================================================
#				UMINUS					      
#				------				      	
# in entry  : 	left  -> a value
#		
#
# in return :   
#
# details   :	It calculates -left by calling the the uminus function of
#		the good class
#
============================================================================*/



value * uminus (value *left, int &err) {
  // calculate -left

  if (left->type == NUM)
    return new value ((-left->num));

  if (left->type == DENSE)
    return new value (new dense (-*left->dmat));

  if (left->type == SPARSE)
    return new value (new sparse (-*left->smat));

  yyerror ("Error, invalid type for operator -");
  err = 1;
  return left;

} // end uminus
    


