// ****************************************************************
//
//      Class string
//
//      String is a class for implementing an abstract data type
//      for  character strings.  It is realy just a simpler 
//      interface to the C string library.
//
//      If you discover any bugs, it would be very helpful if
//      you sent me e-mail.  Thank you.
//
//      Author:                Peter Braaten
//      e-mail:                braa0007@mermaid.micro.umn.edu
//      Date last modified:    July 11, 1994
//
// ****************************************************************

/*==========================================================================
				PRIVATE : 
				--------

    char *ptr;       // pointer to the first character of the string

				FUNCTIONS :
				----------

constructors
destructor

empty :			returns 1 if the string is NULL
get_string(s) : 	copies the value of the string into s
get_string(a,b) : 	returns the string with from the index a to b
length : 		returns the lenght of the string
get_char(i) : 		returns the ith char of the string
set_char(c,i) : 	sets the ith char to c
remove_leading_spaces : removes the spaces from the string up to the first
			character that is not a space or the NULL character
remove_trailing_spaces :removes the spaces from the end of the string up to the
			last character that is not a space

operators : 		=, <<, [], +, <, ==, >, !=, ()

is_letter(c) : 		returns 1 if c is an upper or lower case letter
is_numeral(c) : 	returns 1 if c is numeral
input(st) : 		gets a string from the keyboard (finished by a double
			return)

==========================================================================*/

#include "spmath.h"

//#define debugging
#define line_length 500

//*******************************************************************
//      Member Constructors
//*******************************************************************


string::string () {
  // initialize an empty string.
  ptr = new char[1];
    
  ptr[0] = '\0';
} // end string 

string::string (const char *str) {
  // initialize a string that will contain str.

  ptr = new char[strlen(str)+1];

  strcpy (ptr,str);
} // end string

string::string (const string & str) {
  // initialize a string that will contain str.

  ptr = new char[str.length()+1];

  strcpy (ptr,str.get_pointer());
} // end string

//******************************************************************
//       Member Destructor
//******************************************************************

string::~string () {
  // Destructor for strings. Free any memory that the string may have held.

  //  delete [] ptr;
  //  ptr = NULL;

} // end string destructor


//******************************************************************
//       Member Selectors
//******************************************************************

/*=========================================================================
#				EMPTY :
#				------
# Entries :	nothing 	
#
# Returns : 	1 if the string is empty (ptr[0]='0')
#		0 else
#
=========================================================================*/

int string::empty () const {
  // Boolean function.  Return true (1) if the
  // string is empty, otherwise false (0).

  return (ptr[0] == '\0');
} // end empty

/*=========================================================================
#				GET_PTR :
#				---------
# Entries :	nothing 	
#
# Returns : 	the pointer to the string : ptr
#
=========================================================================*/

char * string::get_pointer () const {
  // Returns the pointer to the string, ptr.

  return ptr;
} // end get_pointer;

/*=========================================================================
#				GET_STRING :
#				-----------
# Entries :	an array of char (string) -> str 	
#
# Returns : 	nothing
#
# Details : 	It copies the value of str into the string pointed to by ptr 
#
=========================================================================*/

void string::get_string (char *str) const {
  // Copy the value of the string pointed to by ptr 
  // into str.

  strcpy (str,ptr);
} // end get_string

/*=========================================================================
#				GET_STRING :
#				-----------
# Entries :	2 integers :a, b -> the indecis 	
#
# Returns : 	the string from the index a to b
#
# Details :	- it checks if the 2 indecis don't exceed the dimension of
#		  the string
#		- it returns the address of the a'th character of the string
#
=========================================================================*/


string string::get_string (int a, int b) const {
  // Return the string from the index a to b.

  string temp = string(ptr);
  // int i;

  #ifdef debugging
    assert (a >= 0   &&   a <= length());
    assert (b >= 0   &&   b <= length());
  #endif

  if (b <= length() -1)
    temp[b+1] = '\0';


  return string(&temp[a]);

} // end get_string

/*=========================================================================
#				LENGTH :
#				--------
# Entries :	nothing 	
#
# Returns : 	the length of the string
#
=========================================================================*/

int string::length ()  const {
  // Return the length of the string.
  
  // Pete, this is slow.
  return strlen (ptr);
} // end length

/*=========================================================================
#				GET_CHAR :
#				----------
# Entries :	i -> the indecis of the char to get 	
#
# Returns : 	the i'th character of the string (and checks to make sure that 
#		i doesn't exceed the length of the string)
#
=========================================================================*/

char string::get_char (int i) const {
  // Return the ith character in the string.  The 
  // first character is at i = 0.

  #ifdef debugging
    assert (i >= 0   &&   i <= length());
  #endif

  return ptr[i];
} // end get_char


int string::size() {
  return (sizeof(char)*strlen(ptr) +
	  sizeof(*this));
}


//******************************************************************
//       Member Functions
//******************************************************************

/*=========================================================================
#				SET_CHAR :
#				----------
# Entries :	- c -> a char
# 		- i -> an integer (the index)
#
# Returns : 	nothing
#
# Details :	- It checks if the index doesn't exceed the length of the string
#		- It copies c into ptr[i]
#
=========================================================================*/

void string::set_char (char c, int i) {
  // Set the ith character to c.

  #ifdef debugging
    assert (i >= 0   &&   i <= length());
  #endif

  ptr[i] = c;
} // end set_char

/*=========================================================================
#			REMOVE_LEADING_SPACES :
#			-----------------------
# Entries :	nothing 	
#
# Returns : 	nothing
#
# Details :	It removes the spaces from the string up to the first 
#		character that is not a space
#
=========================================================================*/

void string::remove_leading_spaces () {
  // Remove the spaces from the string up to the
  // first character that is not a space, or the 
  // null character.

  while (*ptr == ' ') {
    ptr += 1;
  }

} // end  remove_leading_spaces

/*=========================================================================
#			REMOVE_TRAILING_SPACES :
#			-----------------------
# Entries :	nothing 	
#
# Returns : 	nothing
#
# Details :	It removes the spaces from the string up to the last 
#		character that is not a space
#		- It goes at the end of the string
#		- Each space becomes a '\0' (end of string)
#
=========================================================================*/

void string::remove_trailing_spaces () {
  // Remove the spaces from the end of the string up
  // to the last character that is not a space, or the
  // first character is reached (the string becomes null).

  int i = 0;

  while (*(ptr+i) != '\0') 
    i += 1;

  if (i != 0)
    i -= 1;

  while (*(ptr+i) == ' '   &&   i >= 0) {
    *(ptr+i) = '\0';
    i -= 1;
  } // end while

} // end remove_trailing_spaces

//******************************************************************
//     Member Operators
//******************************************************************

/*=========================================================================
#				OPERATOR = :
#				------------
# Entries :	a string y 	
#
# Returns : 	a string : result
#
# Details :	It copies the right string (y) into the left string (instance)
#		and returns the value
#
=========================================================================*/

string string::operator = (const string & y) {
  // Copy the right string (y) into the left string (this instance).
  // Return the value.

  delete [] ptr;
  ptr = new char [y.length() + 1];
  
  strcpy (ptr,y.get_pointer());
  
  return y;
} // end operator =

/*=========================================================================
#				OPERATOR <<= :
#				--------------
# Entries :	a string y 	
#
# Returns : 	a string : result
#
# Details :	It copies the pointer to the right string (y) into the 
#		pointer to the left string (instance) and returns the value
#
=========================================================================*/

string string::operator <<= (string & y) {
  // Copy the right string (y) into the left string (this instance).
  // Return the value.  The actual buffer is not copied.

  delete [] ptr;

  ptr = y.get_pointer();
  
  return y;
} // end operator <<

/*=========================================================================
#				OPERATOR [] :
#				-------------
# Entries :	i -> an integer (the index) 	
#
# Returns : 	a character : result
#
# Details :	It returns the i'th character of the string (see get_char)
#
=========================================================================*/

char& string::operator[] (int i) {
  // Return the ith element by reference.  This routine is
  // paraphrased from "C++ Primer" by Lippman, p 309.

  #ifdef debugging
    assert (i >= 0   &&   i <= length());
  #endif

  return ptr[i];
} // end operator []

/*=========================================================================
#				OPERATOR * :
#				------------
# Entries :	nothing 	
#
# Returns : 	a string : result
#
# Details :	It converts a string from this class into a normal string
#		pointer
#
=========================================================================*/


string::operator char * () {
  // Conversion operator to convert to a normal
  // string pointer

  return ptr;
} // end char *

//******************************************************************
//     Operators
//******************************************************************

/*=========================================================================
#				OPERATOR + :
#				------------
# Entries :	a string x
#		a string y 	
#
# Returns : 	a string : result = concatenation of x and y
#
# Details :	- it calculates the length of the new string
#		- it copies x into this new string
#		- It concatenates y at the end of x
#
=========================================================================*/

string operator + (const string & x, const string & y) {
  // Concatanate two strings together and return it.

  char *str = new char [x.length() + y.length() + 1];

  strcpy (str,x.get_pointer());
  strcat (str,y.get_pointer());

  return string (str);
} // end operator +

/*=========================================================================
#				OPERATOR < :
#				------------
# Entries :	a string x
#		a string y 	
#
# Returns : 	an integer -> result
#
# Details :	it returns 1 if x<y (alphabetically)
#
=========================================================================*/

int operator < (const string & x, const string & y) {
  // Boolean  function:  return true if x < y (alphabetically).

  return (strcmp (x.get_pointer(), y.get_pointer()) < 0);
} // end operator < 

/*=========================================================================
#				OPERATOR == :
#				------------
# Entries :	a string x
#		a string y 	
#
# Returns : 	an integer -> result
#
# Details :	it returns 1 if x==y (alphabetically)
#
=========================================================================*/

int operator == (const string & x, const string & y) {
  // Boolean  function:  return true if x == y (alphabetically).

  return (strcmp (x.get_pointer(), y.get_pointer()) == 0);
} // end operator == 

/*=========================================================================
#				OPERATOR > :
#				------------
# Entries :	a string x
#		a string y 	
#
# Returns : 	an integer -> result
#
# Details :	it returns 1 if x>y (alphabetically)
#
=========================================================================*/

int operator > (const string & x, const string & y) {
  // Boolean  function:  return true if x > y (alphabetically).

  return (strcmp (x.get_pointer(), y.get_pointer()) > 0);
} // end operator > 

/*=========================================================================
#				OPERATOR != :
#				-------------
# Entries :	a string x
#		a string y 	
#
# Returns : 	an integer -> result
#
# Details :	it returns 1 if x!=y 
#
=========================================================================*/

int operator != (const string &x, const string & y) {
  //  Boolean function:  return true if x != y.

  return (!(x == y));
} // end operator !=

//******************************************************************
//     Functions
//******************************************************************
/*=========================================================================
#			REMOVE_LEADING_SPACES :
#			-----------------------
# Entries :	str -> a string 	
#
# Returns : 	a new string
#
# Details :	It removes the spaces from the string up to the first 
#		character that is not a space
#
=========================================================================*/

string remove_leading_spaces (string & str) {
  // Return a string the same as str but with
  // the leading spaces removed.

  string temp = str;

  temp.remove_leading_spaces();
  
  return temp;
} // end remove_leading_spaces

/*=========================================================================
#			REMOVE_TRAILING_SPACES :
#			-----------------------
# Entries :	str -> a string 	
#
# Returns : 	a new string
#
# Details :	see remove_trailing_spaces()
#
=========================================================================*/

string remove_trailing_spaces (string & str) {
  // Return a string the same as str but with
  // the trailing spaces removed.

  string temp = str;

  temp.remove_trailing_spaces();
  
  return temp;
} // end remove_trailing_spaces

/*=========================================================================
#				OUTPUT :			
#				-------
# Entries :	str -> a string 	
#
# Returns : 	nothing
#
# Details :	It prints out the value of the string
#
=========================================================================*/

void output (const string & str) {
  // Print out the value of the string.

  sp_out(str.get_pointer());
} // end output

/*=========================================================================
#				IS_LETTER :
#				----------
# Entries :	x -> a characater 	
#
# Returns : 	an integer (1 if x is an upper or a lower case)
#
=========================================================================*/

int is_letter (char x) {
  // Boolean function.  Return true if x is an upper,
  // or lower case letter, otherwise false.

  return ((x >= 'a' && x <= 'z')   ||   (x >= 'A' && x <= 'Z'));
} // end is_letter

/*=========================================================================
#				IS_NUMERAL :
#				-----------
# Entries :	x -> a characater 	
#
# Returns : 	an integer (1 if x is a numeral)
#
=========================================================================*/

int is_numeral (char x) {
  // Boolean function.  Return true if x is a numeral,
  // otherwise false.

  return (x >= '0' && x <= '9');
} // end is_numeral

/*=========================================================================
#				INPUT :
#				------
# Entries :	result -> a string 	
#
# Returns : 	a string
#
# Details :	It gets a string from the keyboard (ended by a double return)
#		- It creates an array of char of the length of the line
#		- It gets the line and ends the char by a "\n\0"
#		- it concatenates result with this line
#
=========================================================================*/

string input (string & result) {
  // Get a string from the keyboard.  A double return 
  // signifies that the user is finished.  All other
  // return characters are replaced with spaces.
  // The length of the string can be any length, however
  // each line can not exceed the size of the array
  // exp.
  
  char exp[line_length];   // Character array to hold each line. 
  int length = -1;         // Int for the length of exp
  int i;                   // Simple loop index.

  result = "";    

  do {  
    cin.getline (exp, line_length, '\n');  // Get a line.
    length = strlen(exp);                  // Determine its length.
    exp[length] = '\n';                    // Replace its null with \n
    exp[length+1] = '\0';                  // Append a null.
    result = result + exp;                 // Append line to the string.
  }
  while (result.length() <= 1 || result[result.length()-2] !='\n');
 // Keep going until a double \n

  // Replace the return characters with spaces.
  int len = result.length();
  for (i = 0; i <= len -1; i++)
    if (result[i] == '\n'  ||  result[i] == '\t')
      result[i] = ' ';

  return result;

} // end input

